"""
    Tests to validate portfolio metrics
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import unittest

from blankly.metrics import *


class Metrics(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.start_value = 100
        cls.end_value = 150
        cls.years = 2
        cls.returns = [0.03, 0.05, -0.06, -0.02, 0.03, 0.025, 0.5, 0]

    def test_cagr(self):
        truth = 0.2247448714
        result = cagr(self.start_value, self.end_value, self.years)
        self.assertAlmostEqual(truth, result)
    
    def test_cum_returns(self):
        value = cum_returns(self.start_value, self.end_value)
        self.assertEqual(0.5, value)
    
    def test_sortino(self):
        truth = 38.93662218111633
        result = sortino(self.returns)
        self.assertEqual(truth, result)

    def test_sharpe(self):
        truth = 6.206188187237863
        result = sharpe(self.returns)
        self.assertEqual(truth, result)

    def test_maxdrawdown(self):
        truth = -0.07880
        result = max_drawdown(self.returns)
        self.assertAlmostEqual(truth, result)


"""
    Class for defining headers for deployment infrastructure
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import smtplib
import ssl

from typing import Any

from blankly.utils.utils import load_notify_preferences
from blankly.frameworks.strategy import Strategy
from blankly.frameworks.screener.screener import Screener
from blankly.exchanges.interfaces.paper_trade.backtest_result import BacktestResult


class Reporter:
    def __init__(self):
        self.__live_vars = {}
        self.__screener = None
        pass

    def export_live_var(self, var: Any, name: str, description: str = None):
        """
        Create a variable that can be updated by external processes
        All strings must be in ascii characters

        Args:
            var: Any variable that can represented in a string (ex: float, str, int)
            name: The name of the live_var
            description (optional): A longer description for use in GUIs or other areas where context is important
        """
        self.__live_vars[id(var)] = var

    def update_live_var(self, var):
        """
        Get the variable as with any changes that may have occurred

        Args:
            var: The variable that was exported initially
        """
        return self.__live_vars[id(var)]

    def export_strategy(self, strategy: Strategy):
        """
        Export a strategy for monitoring. This is used internally on the construction of the strategy object

        Args:
            strategy (Strategy): The strategy object to monitor
        """
        pass

    def export_screener(self, screener: Screener):
        """
        Export a screener object to the backend for monitoring

        Args:
            screener: A screener object to export
        """
        if self.__screener is None:
            self.__screener = screener
        else:
            raise RuntimeError("Currently only a single screener can be exported per model.")

    def export_screener_result(self, screener: Screener):
        """
        Re-export for the finished screener result

        Args:
            screener: A screener object to export
        """
        pass

    def export_backtest_result(self, backtest_result: BacktestResult):
        """
        Re-export for the finished screener result

        Args:
            backtest_result: A blankly backtest result object
        """
        pass

    def export_used_exchange(self, exchange_name):
        """
        Export the string identifier for a used exchange
        """
        pass

    def log_market_order(self, exchange_out: dict, exchange_type: str):
        """
        Log a market order
        """
        pass

    def log_limit_order(self, exchange_out: dict, exchange_type: str):
        """
        Log a limit order
        """
        pass

    def update_order(self, exchange_out: dict, exchange_type: str):
        """
        Update an existing order
        """
        pass

    def export_used_symbol(self, symbol):
        """
        Export the currency name for a used currency
        """
        pass

    def annotate_order(self, order_id, annotation):
        """
        Update the annotation on an existing order
        """
        pass

    def text(self, text: str):
        """
        Send a text message to the number if notify.json OR the phone number attached to your account if the model
        is deployed live

        Args:
            text: The message body to be sent to your phone number
        """
        notify_preferences = load_notify_preferences()
        provider = notify_preferences['text']['provider']
        phone_number = notify_preferences['text']['phone_number']

        email_lookup = {
            'att': '@txt.att.net',
            'boost': '@smsmyboostmobile.com',
            'cricket': '@sms.cricketwireless.net',
            'sprint': '@messaging.sprintpcs.com',
            't_mobile': '@tmomail.net',
            'us_cellular': '@email.uscc.net',
            'verizon': '@vtext.com',
            'virgin_mobile': '@vmobl.com'
        }
        try:
            email = email_lookup[provider]
        except KeyError:
            raise KeyError("Provider not found. Check the notify.json documentation to see supported providers.")

        self.__send_email(text, override_receiver=phone_number + email)

    @staticmethod
    def __send_email(email_str: str, override_receiver=None):
        """
        Internal email send. This is separated because override_receiver shouldn't be exposed to the user
        """
        notify_preferences = load_notify_preferences()
        port = notify_preferences['email']['port']
        smtp_server = notify_preferences['email']['smtp_server']
        sender_email = notify_preferences['email']['sender_email']
        receiver_email = notify_preferences['email']['receiver_email']
        password = notify_preferences['email']['password']

        if override_receiver is not None:
            receiver_email = override_receiver

        message = email_str

        context = ssl.create_default_context()
        with smtplib.SMTP_SSL(smtp_server, port, context=context) as server:
            server.login(sender_email, password)
            server.sendmail(sender_email, receiver_email, message)

    def email(self, email: str):
        """
        Send an email to the email specified in the notify.json OR your user account email if run live

        Args:
            email: The body of the email to send
        """
        self.__send_email(email)



"""
    Custom API for interacting with the frontend API routes
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json

import requests
from blankly.utils.utils import info_print

blankly_deployment_url = 'https://deploy.blankly.finance'


class API:
    def __init__(self, token, override_url: str = None):
        if override_url:
            self.url = override_url
        else:
            self.url = blankly_deployment_url

        self.token = None
        self.auth_data = self.exchange_token(token)

        try:
            self.token = self.auth_data['idToken']
        except (KeyError, TypeError):
            raise KeyError("Failed to authenticate - run \"blankly login\" again.")
        self.user_id = self.auth_data['data']['user_id']

    def __request(self, type_: str, route: str, json_: dict = None, params: dict = None, file=None, data: dict = None):
        """
        Create a general request to the blankly API services

        Args:
            type_: Request types such as 'post', 'get', and 'delete'
            route: The address where the request should be routed to './model/details'
            json_: Optional json to be attached to the request
            params: Optional parameters for the address URL
            data: Optional JSON to be attached to the request body dictionary
            file: Optional file uploaded in bytes: file = {'file': open(file_path, 'rb')}
        """
        url = self.url
        if url[-1] != '/' and route[0] != '/':
            url += '/'
        url += route

        kwargs = {
            'url': url,
            'params': params,
            'json': json_,
            'files': file,
            'data': data,
        }

        # Add the token if we have it
        if self.token is not None:
            kwargs['headers'] = {'token': self.token}

        try:
            if type_ == "get":
                out = requests.get(**kwargs)
            elif type_ == "post":
                out = requests.post(**kwargs)
            elif type_ == "delete":
                out = requests.delete(**kwargs)
            else:
                raise LookupError("Request type is not implemented or does not exist.")
        except requests.exceptions.ConnectionError:
            raise requests.exceptions.ConnectionError("Failed to connect to deployment service.")

        # Show info that the request was not authorized but still
        #  allow the process to continue
        if out.status_code == 401:
            info_print("Unauthorized request.")

        try:
            return out.json()
        except ValueError:
            print(f'Invalid Response: {out}')
            return None

    def exchange_token(self, token):
        """
        Get the JWT from the refresh token
        """
        return self.__request('post', 'auth/token', data={'refreshToken': token})

    def get_details(self, model_id: str):
        """
        Get the details route
        """
        return self.__request('post', 'model/details', data={'modelId': model_id})

    def get_status(self):
        return self.__request('get', 'model/status')

    def list_projects(self):
        return self.__request('get', 'project/list')

    def get_plans(self, type_: str):
        """
        Args:
            type_: Can be 'backtesting' or 'live'
        """
        return self.__request('post', 'project/plans', data={'type': type_})

    def create_project(self, name: str, description: str):
        return self.__request('post', 'project/create', data={'name': name,
                                                              'description': description})

    def deploy(self, file_path: str, model_id: str, version_description: str,
               python_version: float, type_: str, plan: str, schedule: str = None):
        file_path = r'{}'.format(file_path)
        file = {'model': open(file_path, 'rb')}
        return self.__request('post', 'model/deploy', file=file, data={'pythonVersion': python_version,
                                                                       'versionDescription': version_description,
                                                                       'projectId': self.user_id,
                                                                       'modelId': model_id,
                                                                       'type': type_,
                                                                       'plan': plan,
                                                                       'schedule': schedule})

    def backtest_deployed(self, model_id: str, args: dict, version_id: str, backtest_description: str):
        return self.__request('post', 'model/backtestUploadedModel',
                              json_={'projectId': self.user_id,
                                     'modelId': model_id,
                                     'versionId': version_id,
                                     'backtestArgs': args,
                                     'backtestDescription': backtest_description})

    def backtest(self, file_path: str, model_id: str, args: dict, plan: str,
                 type_: str, python_version: float, backtest_description: str = ""):
        file_path = r'{}'.format(file_path)
        file = {'model': open(file_path, 'rb')}
        return self.__request('post', 'model/backtest', file=file,
                              data={'pythonVersion': str(python_version),
                                    'projectId': self.user_id,
                                    'modelId': model_id,
                                    'type': type_,
                                    'backtestArgs': json.dumps(args),
                                    'backtestDescription': backtest_description,
                                    'plan': plan,
                                    })

    def create_model(self, type_: str, name: str, description: str):
        return self.__request('post', 'model/create-model',
                              data={
                                  'projectId': self.user_id,
                                  'type': type_,
                                  'name': name,
                                  'description': description
                              })

    def list_models(self):
        return self.__request('post', 'model/list',
                              data={
                                  'projectId': self.user_id
                              })

    def generate_keys(self):
        return self.__request('post', 'project/generate-project-token', data={
            'projectId': self.user_id
        })



"""
    Wrapper code for creating
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import time


class BacktestingWrapper:
    def __init__(self):
        self.backtesting = False
        self.frame = {
            'prices': {},
            'time': 0
        }

    def set_backtesting(self, status: bool):
        self.backtesting = status

    def receive_time(self, new_time):
        self.frame['time'] = new_time

    def receive_price(self, asset_id, new_price):
        self.frame['prices'][asset_id] = new_price

    """
    Override functions for manipulating backtesting
    """

    def get_backtesting_price(self, asset_id):
        try:
            return self.frame['prices'][asset_id]
        except KeyError:
            raise KeyError(f"Price not found in recent frame. Have prices for {asset_id} been downloaded?")

    def time(self):
        if self.backtesting:
            return self.frame['time']
        else:
            return time.time()



"""
    Indicator utils and repeatedly used functions
    Copyright (C) 2021 Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Any
from collections import deque

import numpy as np
import pandas as pd


def to_historical_returns(data: Any):
    return pd.Series(data).diff().tolist()


def convert_to_numpy(data: Any):
    if isinstance(data, list) or isinstance(data, deque):
        return np.fromiter(data, float)
    elif isinstance(data, pd.Series):
        return data.to_numpy()
    return data


def check_series(data: Any):
    return isinstance(data, pd.Series)



from blankly.exchanges.auth.auth_constructor import AuthConstructor
from blankly.exchanges.futures.futures_exchange import FuturesExchange
from blankly.exchanges.interfaces.ftx.ftx_api import FTXAPI
from blankly.exchanges.interfaces.ftx_futures.ftx_futures_interface import FTXFuturesInterface
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface


class FTXFutures(FuturesExchange):

    def __init__(self,
                 portfolio_name=None,
                 keys_path="keys.json",
                 preferences_path=None):
        super().__init__("ftx_futures", portfolio_name, preferences_path)

        # Load auth from keys file
        auth = AuthConstructor(keys_path, portfolio_name, 'ftx',
                               ['API_KEY', 'API_SECRET', 'sandbox'])

        if auth.keys['sandbox']:
            raise Exception(
                'FTX Futures exchange does not support sandbox mode')

        tld = self.preferences["settings"]["ftx_futures"]["ftx_tld"]
        if tld != 'com':
            raise Exception(
                f'FTX Futures exchange does not support .{tld} tld.')

        self.__calls = FTXAPI(auth.keys['API_KEY'],
                              auth.keys['API_SECRET'],
                              tld=tld)

        self.__interface = FTXFuturesInterface(self.exchange_type, self.calls)

        self.initialize()

    @property
    def calls(self):
        return self.__calls

    @property
    def interface(self) -> FuturesExchangeInterface:
        return self.__interface



"""
    Allows the user to load the Keys in any Keys.json.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
from blankly.utils.utils import info_print

keys_path_cache = None

# This is currently unused
auth_cache = {}


# def load_auth_coinbase_pro(keys_file, name):
#     exchange_type = "coinbase_pro"
#     name, portfolio = __load_auth(keys_file, name, exchange_type)
#
#     return [portfolio["API_KEY"], portfolio["API_SECRET"], portfolio["API_PASS"]], name
#
#
# def load_auth_binance(keys_file, name):
#     exchange_type = "binance"
#     name, portfolio = __load_auth(keys_file, name, exchange_type)
#
#     return [portfolio["API_KEY"], portfolio["API_SECRET"]], name


def load_auth(exchange_type, keys_file=None, name=None):
    global keys_path_cache

    if keys_file is None:
        if keys_path_cache is None:
            # Add a default for if info is passed in and nobody knows anything about paths
            keys_path_cache = './keys.json'
            keys_file = keys_path_cache
        else:
            # Default to the cached path if the passed variable is wrong
            keys_file = keys_path_cache
    else:
        # If it's not none then there's no problem, just write it to the cache though
        keys_path_cache = keys_file

    auth_object = load_json(keys_file)
    try:
        exchange_keys = auth_object[exchange_type]
    except KeyError as e:
        raise KeyError(f"Cannot find exchange {e} the keys file.")
    if name is None:
        name, portfolio = __determine_first_key(exchange_keys)
    else:
        portfolio = exchange_keys[name]

    if exchange_type not in auth_cache:
        auth_cache[exchange_type] = {}
    return name, portfolio


def __determine_first_key(exchange_keys):
    first_key = list(exchange_keys.keys())[0]
    warning_string = "No portfolio name to load specified, defaulting to the first in the file: " \
                     "(" + first_key + "). This is fine if there is only one portfolio in use."
    info_print(warning_string)
    # Read the first in the portfolio
    portfolio = exchange_keys[first_key]
    name = first_key
    return name, portfolio


def write_auth_cache(exchange, name, auth):
    """
    Write an authenticated object into the global authentication cache. This can be used by other pieces of code
    in the module to pull from exchanges at points they need the API

    Args:
        exchange (str): Exchange name ex: "coinbase_pro" or "binance"
        name (str): Portfolio name ex: "my cool portfolio"
        auth (obj): Authenticated object to store & recover.
    """
    global auth_cache
    if exchange not in auth_cache:
        auth_cache[exchange] = {}

    auth_cache[exchange][name] = auth


def read_auth_cache(exchange, name=None):
    """
    Pull an authenticated object on an exchange. This can be used for API calls anywhere in the code.
    """
    global auth_cache
    if exchange in auth_cache:
        if name is None:
            first_key = list(auth_cache[exchange])[0]
            return auth_cache[exchange][first_key]
        else:
            return auth_cache[exchange][name]
    else:
        raise KeyError("Request exchange invalid or does not exist")


def load_json(keys_file):
    try:
        f = open(keys_file)
        contents = json.load(f)
        f.close()
        return contents
    except FileNotFoundError:
        raise FileNotFoundError("Make sure a Keys.json file is placed in the same folder as the project working "
                                "directory!")


def default_first_portfolio(keys_file, exchange_name):
    auth_object = load_json(keys_file)
    exchange_keys = auth_object[exchange_name]
    first_key = list(exchange_keys.keys())[0]
    warning_string = "No portfolio name to load specified, defaulting to the first in the file: " \
                     "(" + first_key + "). This is fine if there is only one portfolio in use."
    info_print(warning_string)
    # Read the first in the portfolio
    portfolio = exchange_keys[first_key]
    name = first_key
    return name, portfolio



"""
    Logging system to allow users to view & understand actions done in the strategy
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from abc import ABC

from blankly.enums import Side
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface

# TODO
from blankly.exchanges.orders.futures.futures_order import FuturesOrder


class FuturesStrategyLogger:
    interface: FuturesExchangeInterface

    def __init__(self, interface=None, strategy=None):
        self.interface = interface
        self.strategy = strategy

    def __getattribute__(self, item):
        # run the overridden function in this class if it exists, or default to the method in self.interface
        try:
            # this is NOT recursive, it will get the *actual* attribute
            return object.__getattribute__(self, item)
        except AttributeError:
            return self.interface.__getattribute__(item)

    # TODO log order (and other order types)
    # example:
    def market_order(self, symbol: str, side: Side, size: float, *args,
                     **kwargs) -> FuturesOrder:
        # log here
        return self.interface.market_order(symbol, side, size, *args, **kwargs)



"""
    Unit tests for binance.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import time
import unittest
from pathlib import Path

import pytest

import blankly

from blankly.exchanges.interfaces.binance.binance_interface import BinanceInterface
from blankly.utils.utils import AttributeDict


class BinanceInterface_test(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.Binance = blankly.Binance(portfolio_name="Spot Test Key",
                                      keys_path="./tests/config/keys.json",
                                      settings_path="./tests/config/settings.json")
        cls.Binance_Interface = cls.Binance.get_interface()

    def test_get_exchange_type(self):
        binance = self.Binance_Interface.get_exchange_type()
        self.assertEqual(binance, "binance")


@pytest.fixture
def binance_interface():
    keys_file_path = Path("tests/config/keys.json").resolve()
    settings_file_path = Path("tests/config/settings.json").resolve()

    binance = blankly.Binance(keys_path=keys_file_path,
                              settings_path=settings_file_path, portfolio_name='Spot Test Key')

    return binance.interface


def test_get_exchange(binance_interface: BinanceInterface) -> None:
    assert binance_interface.get_exchange_type() == 'binance'


def test_get_account(binance_interface: BinanceInterface) -> None:
    resp = binance_interface.get_account()
    assert type(resp) is AttributeDict

    found_btc = False
    for asset in resp.keys():
        if asset == "BTC":
            resp['BTC']['available'] += 1
            resp['BTC']['hold'] += 1

            found_btc = True

    assert found_btc


def test_get_buy_sell_order(binance_interface: BinanceInterface) -> None:
    # query for the unique ID of BTC-USD
    products = binance_interface.get_products()
    btc_usd_id = None

    for i in products:
        if i['symbol'] == 'BTC-USDT':
            btc_usd_id = i['symbol']
            break
    print(btc_usd_id)
    assert btc_usd_id

    market_buy_order = binance_interface.market_order(btc_usd_id, 'buy', .01)

    time.sleep(1.5)

    num_retries = 0
    # wait until order is filled
    while binance_interface.get_order(btc_usd_id, market_buy_order.get_id())['status'] != "filled":
        time.sleep(1)
        num_retries += 1
        if num_retries > 10:
            raise TimeoutError("Too many retries, cannot get order status")

    resp = binance_interface.get_order(btc_usd_id, market_buy_order.get_id())
    assert type(resp) is dict
    assert resp["side"] == "buy"
    assert resp["type"] == "market"

    market_sell_order = binance_interface.market_order(btc_usd_id, 'sell', .01)
    # this test has been flaky, so lets add a sleep for engine to process order
    time.sleep(1)
    # wait until order is filled
    while binance_interface.get_order(btc_usd_id, market_sell_order.get_id())['status'] != "filled":
        time.sleep(1)
        num_retries += 1
        if num_retries > 10:
            raise TimeoutError("Too many retries, cannot get order status")

    resp = binance_interface.get_order(btc_usd_id, market_sell_order.get_id())
    assert type(resp) is dict
    assert resp["side"] == "sell"
    assert resp["type"] == "market"


def test_get_price(binance_interface: BinanceInterface) -> None:
    # todo: just call a couple of popular coins and see if they all return floats
    btcusd = 'BTC-USDT'
    resp = binance_interface.get_price(btcusd)
    assert type(resp) is float


"""
    This shows how you can easily wrap in your own logic. Pull in your own scripts to make them backtestable and make
    them run live
"""

from blankly import Strategy, StrategyState, Interface
from blankly import Alpaca

# Example import (these imports will fail they just show how you might call a module)
from models import OrderPricingModel, OrderDecisionModel


def init(symbol, state: StrategyState):
    # initialize this once and store it into state
    variables = state.variables
    variables['decision_model'] = OrderDecisionModel(symbol)
    variables['pricing_model'] = OrderPricingModel(symbol)
    variables['has_bought'] = False


def price_event(price, symbol, state: StrategyState):
    interface: Interface = state.interface
    variables = state.variables
    decision_model = variables['decision_model']
    pricing_model = variables['pricing_model']

    # make a decision to buy, sell, or hold
    decision = decision_model(symbol)

    if decision == 0:
        curr_value = interface.account[symbol].available * price
        # call pricing model to determine how much to buy
        amt_to_buy = pricing_model(price, symbol, interface.cash, curr_value)
        interface.market_order(symbol, 'buy', amt_to_buy)
    elif decision == 1:
        curr_value = interface.account[symbol].available * price
        amt_to_sell = pricing_model(price, symbol, interface.cash, curr_value)
        interface.market_order(symbol, 'sell', amt_to_sell)


if __name__ == "__main__":
    alpaca = Alpaca()
    s = Strategy(alpaca)
    s.add_price_event(price_event, 'MSFT', resolution='1d', init=init)
    # decision_model = OrderDecisionModel() <-- global state can also be accessed in price event functions
    # pricing_model = OrderPricingModel()
    s.backtest(initial_values={'USD': 10000}, to='2y')




import sys
import time
from datetime import datetime
from operator import itemgetter

import pytest
from _pytest.python_api import approx

from blankly.enums import Side, OrderType, ContractType, OrderStatus, HedgeMode, MarginType, PositionMode, TimeInForce
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface
from blankly.utils import utils
from tests.new_interface_tests.test_utils import wait_till_filled, homogeneity_testing, sell, buy, cancelling_order, close_position, \
    close_all_positions

# TODO auto truncate
# TODO min size/min notional api


def valid_product_helper(futures_interface: FuturesExchangeInterface, product):
    base = product['base_asset']
    quote = product['quote_asset']
    symbol = product['symbol']
    assert symbol == base + '-' + quote
    exc = futures_interface.to_exchange_symbol(symbol)
    assert futures_interface.to_blankly_symbol(exc) == symbol


def test_valid_symbols(futures_interface: FuturesExchangeInterface):
    for product in futures_interface.get_products().values():
        valid_product_helper(futures_interface, product)


def test_valid_symbol_from_position(
        futures_interface: FuturesExchangeInterface, symbol: str):
    with cancelling_order(futures_interface, symbol):
        products = futures_interface.get_positions().values()
    for product in products:
        valid_product_helper(futures_interface, product)


@homogeneity_testing
def test_get_products(futures_interface: FuturesExchangeInterface):
    return list(futures_interface.get_products().values())


@homogeneity_testing
def test_get_products_symbol(futures_interface: FuturesExchangeInterface,
                             symbol: str):
    return futures_interface.get_products(symbol)


@homogeneity_testing
def test_get_account(futures_interface: FuturesExchangeInterface):
    return list(futures_interface.get_account().values())


@homogeneity_testing
def test_get_positions(futures_interface: FuturesExchangeInterface):
    return list(futures_interface.get_positions().values())


@homogeneity_testing
@pytest.mark.parametrize('side', [Side.BUY, Side.SELL])
def test_get_positions(futures_interface: FuturesExchangeInterface,
                       symbol: str, side: Side):
    if side == Side.BUY:
        func = buy
        m = 1
    elif side == Side.SELL:
        func = sell
        m = -1

    close_position(futures_interface, symbol)

    order = func(futures_interface, symbol)
    position = futures_interface.get_positions(symbol)
    assert position.symbol == symbol
    assert m * position.size == order.size
    assert 0 < position.size * m
    assert position.leverage == futures_interface.get_leverage(symbol)
    # assert position.margin_type == futures_interface.get_margin_type(symbol)

    close_position(futures_interface, symbol)

    return position


def test_simple_open_close(futures_interface: FuturesExchangeInterface,
                           symbol: str) -> None:
    close_position(futures_interface, symbol)

    # buy
    size = buy(futures_interface, symbol).size
    assert futures_interface.get_positions(symbol).size == approx(size)

    close_position(futures_interface, symbol)

    # sell
    size = sell(futures_interface, symbol).size
    assert futures_interface.get_positions(symbol).size == approx(-size)

    close_position(futures_interface, symbol)


@homogeneity_testing
@pytest.mark.parametrize('side', [Side.BUY, Side.SELL])
def test_order(futures_interface: FuturesExchangeInterface, symbol: str,
               side: Side):
    close_position(futures_interface, symbol)

    price = futures_interface.get_price(symbol)
    product = futures_interface.get_products(symbol)
    size = utils.trunc(13 / price, product.size_precision)
    order = futures_interface.market_order(symbol, side, size)

    assert order.symbol == symbol
    assert 0 <= order.id
    assert order.size == approx(size, rel=0.01)
    assert order.status in (OrderStatus.OPEN, OrderStatus.FILLED)
    assert order.type == OrderType.MARKET
    assert order.contract_type == ContractType.PERPETUAL
    assert order.side == side
    assert order.limit_price == 0
    assert order.position == PositionMode.BOTH
    assert order.time_in_force == TimeInForce.GTC

    res = wait_till_filled(futures_interface,
                           order)  # for market orders, just wait for fil

    # make sure we paid what we expect
    assert res.price == approx(size * price, rel=0.01)

    # check our stake is increased/decreased
    position_size = futures_interface.get_positions(symbol).size
    if side == Side.BUY:
        assert order.size == position_size
    else:
        assert order.size == -position_size

    # close position
    close_position(futures_interface, symbol)

    # set price, status and make sure order/res are otherwise equal
    order.price = res.price
    order.status = res.status
    assert order == res

    return order


@homogeneity_testing
@pytest.mark.parametrize(
    'type', [OrderType.LIMIT, OrderType.TAKE_PROFIT, OrderType.STOP])
@pytest.mark.parametrize('side', [Side.BUY, Side.SELL])
def test_priced_orders(futures_interface: FuturesExchangeInterface,
                       symbol: str, side: Side, type: OrderType):
    close_position(futures_interface, symbol)

    price = futures_interface.get_price(symbol)
    product = futures_interface.get_products(symbol)
    size = utils.trunc(20 / price, product.size_precision)
    m = 0.3
    if side == Side.BUY:
        m = -m
    if type == OrderType.STOP:
        m = -m
    limit_price = utils.trunc(price * (1 + m), product.price_precision)
    if type == OrderType.LIMIT:
        order = futures_interface.limit_order(symbol, side, limit_price, size)
    elif type == OrderType.TAKE_PROFIT:
        order = futures_interface.take_profit(symbol, side, limit_price, size)
    elif type == OrderType.STOP:
        order = futures_interface.stop_loss(symbol, side, limit_price, size)

    assert order.symbol == symbol
    assert 0 <= order.id
    assert order.size == approx(size, rel=0.01)
    assert order.status == OrderStatus.OPEN
    assert order.type == type
    assert order.contract_type == ContractType.PERPETUAL
    assert order.side == side
    assert order.limit_price == limit_price
    assert order.position == PositionMode.BOTH
    assert order.time_in_force == TimeInForce.GTC

    res = futures_interface.cancel_order(symbol, order.id)

    order.status = OrderStatus.CANCELED
    assert order == res

    return order


def test_set_hedge_mode(futures_interface: FuturesExchangeInterface):
    close_all_positions(futures_interface)
    if futures_interface.get_exchange_type() == 'ftx_futures':
        pytest.xfail('FTX Futures does not support hedge mode')
    futures_interface.set_hedge_mode(HedgeMode.HEDGE)
    assert futures_interface.get_hedge_mode() == HedgeMode.HEDGE


def test_set_oneway_mode(futures_interface: FuturesExchangeInterface):
    close_all_positions(futures_interface)
    futures_interface.set_hedge_mode(HedgeMode.ONEWAY)
    assert futures_interface.get_hedge_mode() == HedgeMode.ONEWAY


def test_account_leverage(futures_interface: FuturesExchangeInterface):
    close_all_positions(futures_interface)
    if futures_interface.get_exchange_type() == 'binance_futures':
        pytest.xfail(
            'Binance Futures does not support setting account leverage')
    futures_interface.set_leverage(3)
    assert futures_interface.get_leverage() == 3


def test_symbol_leverage(futures_interface: FuturesExchangeInterface,
                         symbol: str):
    if futures_interface.get_exchange_type() == 'ftx_futures':
        close_all_positions(futures_interface)
        futures_interface.set_leverage(3)  # set globally for ftx
        with pytest.raises(
                Exception):  # changing for symbol should raise an exception
            futures_interface.set_leverage(3, symbol)
    else:
        # for other exchanges we can proceed as normal
        close_position(futures_interface, symbol)
        futures_interface.set_leverage(3, symbol)
    assert futures_interface.get_leverage(symbol) == 3


def test_set_cross_margin(futures_interface: FuturesExchangeInterface,
                          symbol: str):
    futures_interface.set_margin_type(symbol, MarginType.CROSSED)

    if futures_interface.get_exchange_type() == 'binance_futures':
        pytest.xfail('Binance doesn\'t have an API for this')
    assert futures_interface.get_margin_type(symbol) == MarginType.CROSSED


def test_set_isolated_margin(futures_interface: FuturesExchangeInterface,
                             symbol: str):
    if futures_interface.get_exchange_type() == 'ftx_futures':
        pytest.xfail('FTX Futures does not support isolated margin')
    futures_interface.set_margin_type(symbol, MarginType.ISOLATED)

    if futures_interface.get_exchange_type() == 'binance_futures':
        pytest.xfail('Binance doesn\'t have an API for this')
    assert futures_interface.get_margin_type(symbol) == MarginType.ISOLATED


@homogeneity_testing
def test_get_open_orders(futures_interface: FuturesExchangeInterface,
                         symbol: str):
    with cancelling_order(futures_interface, symbol) as order:
        open_orders = futures_interface.get_open_orders(symbol)
    assert order in open_orders
    assert all(o.symbol == symbol for o in open_orders)
    return open_orders


@homogeneity_testing
def test_get_order(futures_interface: FuturesExchangeInterface, symbol: str):
    with cancelling_order(futures_interface, symbol) as order:
        assert futures_interface.get_order(symbol, order.id) == order


@homogeneity_testing
def test_price(futures_interface: FuturesExchangeInterface, symbol: str):
    price = futures_interface.get_price(symbol)
    assert 0 < price
    return price


@homogeneity_testing
def test_funding_rate_history(futures_interface: FuturesExchangeInterface,
                              symbol: str):
    day = 60 * 60 * 24
    now = int(datetime.now().timestamp())
    start = now - day * 365
    end = now - day * 7
    history = futures_interface.get_funding_rate_history(symbol=symbol,
                                                         epoch_start=start,
                                                         epoch_stop=end)

    # non-perp contracts don't have funding rates
    if futures_interface.get_products(
    )[symbol].contract_type != ContractType.PERPETUAL:
        assert len(history) == 0
        return

    # test start and end times
    assert start <= history[0]['time'] < start + day
    assert end - day < history[-1]['time'] <= end

    # test ascending order
    assert sorted(history, key=itemgetter('time')) == history

    # test resolution
    errors = []
    resolution = futures_interface.get_funding_rate_resolution()
    for i in range(1, len(history)):
        prev = history[i - 1]
        current = history[i]
        real_res = current['time'] - prev['time']
        if real_res != resolution:
            errors.append(current)
            print(
                f'wrong resolution in funding rate data: {real_res}, should be {resolution}',
                file=sys.stderr)
            if len(errors) > 20:
                pytest.fail(f'too many incorrect resolutions: errors={errors}')
                # pytest.fail(f'too many incorrect resolutions: {errors=}')

    return history



"""
    Base class for FuturesExchangeInterfaces.
    Provides common functionality for Futures Exchanges.
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import abc
from typing import Union, Optional, List

import numpy
import pandas

from datetime import datetime as dt

from dateutil.parser import parser

import blankly.utils.utils as utils
from blankly.enums import MarginType, HedgeMode, PositionMode, OrderType, Side, TimeInForce, ContractType
from blankly.exchanges.interfaces.abc_base_exchange_interface import ABCBaseExchangeInterface
from blankly.exchanges.orders.futures.futures_order import FuturesOrder


class FuturesExchangeInterface(ABCBaseExchangeInterface, abc.ABC):
    exchange_name: str

    def __init__(self,
                 exchange_name,
                 authenticated_api,
                 preferences_path=None):
        self.exchange_name = exchange_name
        self.calls = authenticated_api

        self.user_preferences = utils.load_user_preferences(preferences_path)

        self.exchange_properties = None
        self.available_currencies = {}

        if self.user_preferences['settings']['test_connectivity_on_auth']:
            self.init_exchange()

    @staticmethod
    @abc.abstractmethod
    def to_blankly_symbol(symbol: str):
        pass

    @staticmethod
    @abc.abstractmethod
    def to_exchange_symbol(symbol: str):
        pass

    def get_exchange_type(self) -> str:
        """Returns the exchange type (ex. 'binance', 'coinbase', 'alpaca')"""
        return self.exchange_name

    @abc.abstractmethod
    def init_exchange(self):
        """Initializes the exchange"""
        pass

    @abc.abstractmethod
    def get_products(self, symbol: str = None) -> dict:
        """Returns a list of all products traded on this exchange"""
        pass

    @abc.abstractmethod
    def get_account(self, symbol: str = None) -> utils.AttributeDict:
        """Returns account information, or information for only one `symbol` if one is given."""
        pass

    # TODO this metohd name might need to change to get_position ?
    @abc.abstractmethod
    def get_positions(self, symbol: str = None) -> Optional[dict]:
        """Returns position information, or information for only one `symbol` if one is given"""
        pass

    @abc.abstractmethod
    def market_order(self,
                     symbol: str,
                     side: Side,
                     size: float,
                     position: PositionMode = None,
                     reduce_only: bool = None) -> FuturesOrder:
        """Places a market order"""
        pass

    @abc.abstractmethod
    def limit_order(self,
                    symbol: str,
                    side: Side,
                    price: float,
                    size: float,
                    position: PositionMode = None,
                    reduce_only: bool = None,
                    time_in_force: TimeInForce = None) -> FuturesOrder:
        """Places a limit order"""
        pass

    @abc.abstractmethod
    def take_profit(self,
                    symbol: str,
                    side: Side,
                    price: float,
                    size: float,
                    position: PositionMode = None) -> FuturesOrder:
        """Place a take-profit order for a position"""
        pass

    @abc.abstractmethod
    def stop_loss(self,
                  symbol: str,
                  side: Side,
                  price: float,
                  size: float,
                  position: PositionMode = None) -> FuturesOrder:
        """Place a stop-loss order for a position"""
        pass

    @abc.abstractmethod
    def set_hedge_mode(self, hedge_mode: HedgeMode):
        pass

    @abc.abstractmethod
    def get_hedge_mode(self):
        pass

    @abc.abstractmethod
    def set_leverage(self, leverage: int, symbol: str = None):
        pass

    @abc.abstractmethod
    def get_leverage(self, symbol: str = None) -> float:
        pass

    @abc.abstractmethod
    def set_margin_type(self, symbol: str, type: MarginType):
        pass

    @abc.abstractmethod
    def get_margin_type(self, symbol: str):
        pass

    @abc.abstractmethod
    def cancel_order(self, symbol: str, order_id: int) -> FuturesOrder:
        """Cancels an order"""
        pass

    @abc.abstractmethod
    def get_open_orders(self, symbol: str) -> List[FuturesOrder]:
        """Returns all currently open orders, filtered by `symbol` if one is provided."""
        pass

    @abc.abstractmethod
    def get_order(self, symbol: str, order_id: int) -> FuturesOrder:
        """Returns information for the order corresponding to `order_id`"""
        pass

    @abc.abstractmethod
    def get_price(self, symbol: str) -> float:
        """Returns the current price of an asset"""
        pass

    @property
    def account(self) -> utils.AttributeDict:
        """Account information"""
        return self.get_account()

    @property
    def positions(self) -> utils.AttributeDict:
        """Position information"""
        return self.get_positions()

    @property
    def orders(self) -> list:
        """All currently open orders"""
        return self.get_open_orders()

    @property
    def cash(self) -> float:
        """The amount of cash in a portfolio. The currency for 'cash' is set in settings.json"""
        using_setting = self.user_preferences['settings'][
            self.exchange_name]['cash']
        return self.get_account(using_setting)['available']

    @abc.abstractmethod
    def get_funding_rate_history(self, symbol: str, epoch_start: int,
                                 epoch_stop: int) -> list:
        """
        Get the funding rate history between `epoch_start` and `epoch_end`.
        Returns a list of {'rate': int, 'time': int}
        """
        pass

    @abc.abstractmethod
    def get_funding_rate_resolution(self) -> int:
        pass



"""
    __init__ file to give the module access to the libraries.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import blankly.utils.utils
from blankly.exchanges.interfaces.coinbase_pro.coinbase_pro import CoinbasePro
from blankly.exchanges.interfaces.binance.binance import Binance
from blankly.exchanges.interfaces.alpaca.alpaca import Alpaca
from blankly.exchanges.interfaces.oanda.oanda import Oanda
from blankly.exchanges.interfaces.kucoin.kucoin import Kucoin
from blankly.exchanges.interfaces.ftx.ftx import FTX
from blankly.exchanges.interfaces.paper_trade.paper_trade import PaperTrade
from blankly.exchanges.interfaces.keyless.keyless import KeylessExchange
from blankly.frameworks.strategy import Strategy as Strategy
from blankly.frameworks.strategy import StrategyState as StrategyState
from blankly.frameworks.screener.screener import Screener
from blankly.frameworks.screener.screener_state import ScreenerState

from blankly.exchanges.managers.ticker_manager import TickerManager
from blankly.exchanges.managers.orderbook_manager import OrderbookManager
from blankly.exchanges.managers.general_stream_manager import GeneralManager
from blankly.exchanges.interfaces.abc_exchange_interface import ABCExchangeInterface as Interface
from blankly.frameworks.multiprocessing.blankly_bot import BlanklyBot
from blankly.utils.utils import trunc
import blankly.utils.utils as utils
from blankly.utils.scheduler import Scheduler
import blankly.indicators as indicators
from blankly.utils import time_builder

from blankly.enums import Side, OrderType, OrderStatus, TimeInForce

from blankly.deployment.reporter_headers import Reporter as __Reporter_Headers
is_deployed = False
_screener_runner = None

_backtesting = blankly.utils.check_backtesting()
try:
    from blankly_external import Reporter as __Reporter
    reporter = __Reporter
    is_deployed = True
except ImportError:
    reporter = __Reporter_Headers()



"""
    Simple utility functions that enable more robust testing
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import datetime


def get_valid_symbol(exchange: str):
    if exchange == 'binance':
        return 'BTC-USDT'
    elif exchange == 'coinbase_pro':
        return 'BTC-USD'
    elif exchange == 'alpaca':
        return 'AAPL'
    elif exchange == 'oanda':
        return 'EUR-USD'
    elif exchange == 'ftx':
        return 'BTC-USD'
    elif exchange == 'kucoin':
        return 'ETH-USDT'
    else:
        raise LookupError("Specified exchange not found.")


def forex_market_open():
    now = datetime.datetime.now()
    # friday == 4
    # saturday == 5
    # TODO Technically it resumes at 5pm sunday but I don't handle that
    if now.weekday() >= 5:
        return False
    elif now.weekday() >= 4 and now.hour >= 17:
        return False
    else:
        return True


"""
    CLI for interacting with and uploading blankly models.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import argparse
import sys
import traceback
import warnings
import os
import platform
import runpy
import time
import requests
import json
import zipfile
import tempfile
import webbrowser

from blankly.deployment.api import API
from blankly.utils.utils import load_json_file, info_print, load_deployment_settings

very_important_string = """
██████╗ ██╗      █████╗ ███╗   ██╗██╗  ██╗██╗  ██╗   ██╗    ███████╗██╗███╗   ██╗ █████╗ ███╗   ██╗ ██████╗███████╗
██╔══██╗██║     ██╔══██╗████╗  ██║██║ ██╔╝██║  ╚██╗ ██╔╝    ██╔════╝██║████╗  ██║██╔══██╗████╗  ██║██╔════╝██╔════╝
██████╔╝██║     ███████║██╔██╗ ██║█████╔╝ ██║   ╚████╔╝     █████╗  ██║██╔██╗ ██║███████║██╔██╗ ██║██║     █████╗  
██╔══██╗██║     ██╔══██║██║╚██╗██║██╔═██╗ ██║    ╚██╔╝      ██╔══╝  ██║██║╚██╗██║██╔══██║██║╚██╗██║██║     ██╔══╝  
██████╔╝███████╗██║  ██║██║ ╚████║██║  ██╗███████╗██║       ██║     ██║██║ ╚████║██║  ██║██║ ╚████║╚██████╗███████╗
╚═════╝ ╚══════╝╚═╝  ╚═╝╚═╝  ╚═══╝╚═╝  ╚═╝╚══════╝╚═╝       ╚═╝     ╚═╝╚═╝  ╚═══╝╚═╝  ╚═╝╚═╝  ╚═══╝ ╚═════╝╚══════╝
"""

deployment_script_name = 'blankly.json'


# From blender build scripts found at
# https://stackoverflow.com/a/287944/8087739
class TermColors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


supported_exchanges = ['binance.com', 'binance.us',
                       'coinbase_pro', 'alpaca', 'ftx', 'oanda']


def choose_option(choice: str, options: list, descriptions: list):
    """
    Generalized CLI tool to choose between given options

    Args:
        choice: String for "Choose a <choice>" shown to the user
        options: List of string options allowed ex: ['micro', 'small', 'medium']
        descriptions: Description to show for each option corresponding to the earlier options: ex: ['
            miocro:
                CPU: 5,
                RAM: 10
        ']
    """

    def print_descriptions(descriptions_: list, show_index: bool):
        index_ = 0
        for j in descriptions_:
            if show_index:
                j = TermColors.ENDC + str(index_) + ": " + j
            print(j)
            index_ += 1

    try:
        import fcntl
        import termios
        fd = sys.stdin.fileno()

        largest_option_name = 0
        for i in options:
            if len(i) > largest_option_name:
                largest_option_name = len(i)

        oldterm = termios.tcgetattr(fd)
        newattr = termios.tcgetattr(fd)

        def print_selection_index(index_):
            # We pass none when they haven't selected yet
            if index_ is None:
                chosen_plan = "None"
            # When its given a number it will index and find the lengths
            else:
                if index_ < 0:
                    index_ = len(options) - 1
                elif index_ >= len(options):
                    index_ = 0
                chosen_plan = options[index_]

            # Carry it in this string
            string_ = "\r" + TermColors.UNDERLINE + TermColors.OKCYAN + "You have chosen: " + \
                      TermColors.ENDC + " " + TermColors.BOLD + TermColors.OKBLUE + chosen_plan

            remaining_chars = largest_option_name - len(chosen_plan)
            string_ += " " * remaining_chars

            # Final print and flush the buffer
            sys.stdout.write(string_)
            sys.stdout.flush()

            return index_

        print(TermColors.BOLD + TermColors.WARNING + f"Choose a {choice}: " + TermColors.ENDC +
              TermColors.UNDERLINE + "(Use your arrow keys ← →)" + TermColors.ENDC)

        # Print everything out with no index
        print_descriptions(descriptions, False)

        newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
        termios.tcsetattr(fd, termios.TCSANOW, newattr)

        oldflags = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, oldflags | os.O_NONBLOCK)

        index = 0
        print_selection_index(None)
        try:
            while True:
                try:
                    c = sys.stdin.read(1)
                    if c == '[':
                        c = sys.stdin.read(1)
                        if c == 'C':
                            index += 1
                            index = print_selection_index(index)
                        elif c == 'D':
                            index -= 1
                            index = print_selection_index(index)
                    elif c == '\n':
                        break
                except IOError:
                    pass
        finally:
            termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)
            fcntl.fcntl(fd, fcntl.F_SETFL, oldflags)

        print('\n' + TermColors.BOLD + TermColors.WARNING + f"Chose {choice}:" + TermColors.ENDC + " " +
              TermColors.BOLD + TermColors.OKBLUE + options[index] + TermColors.ENDC)

        return options[index]
    except Exception:
        # info_print("Using non-interactive selection on this device.")

        print_descriptions(descriptions, True)

        print(TermColors.BOLD + TermColors.WARNING + f"Choose a {choice}: " + TermColors.ENDC +
              TermColors.UNDERLINE + "(Input the index of your selection)" + TermColors.ENDC)

        index = int(input(TermColors.UNDERLINE + TermColors.OKCYAN +
                    "You have chosen:" + TermColors.ENDC + " "))

        if index < 0 or index > len(options) - 1:
            raise LookupError(f"The index you chose is out of bounds, choose an index between {0} and "
                              f"{len(options) -1}")

        print('\n' + TermColors.BOLD + TermColors.WARNING + f"Chose {choice}:" + TermColors.ENDC + " " +
              TermColors.BOLD + TermColors.OKBLUE + options[index] + TermColors.ENDC)

        return options[index]


def get_project_model_and_name(args, api: API):
    if 'path' not in args or args['path'] is None:
        print(TermColors.WARNING + "Warning - No filepath specified. Assuming the current directory (./)\n" +
              TermColors.ENDC)

        args['path'] = './'

    try:
        f = open(os.path.join(args['path'], deployment_script_name))
        deployment_options = json.load(f)
        f.close()

        # Use this variable to always write if something has changed
        queue_write = False
        """
        This handles identifying the project on login
        """
        if 'model_id' not in deployment_options or \
                'type' not in deployment_options or \
                'model_name' not in deployment_options:
            # This performs all the querying necessary to send the data up
            choice = choose_option('way to connect to a model', ['Create New', 'Choose from Existing'],
                                   ['Generate a new model',
                                    'Attach to an existing model'])

            if choice == 'Create New':
                model_name = input(TermColors.BOLD + TermColors.WARNING +
                                   "Enter a name for your model: " + TermColors.ENDC)

                # Now we know to go ahead and create a new model on the server
                general_description = input(TermColors.BOLD + TermColors.WARNING +
                                            "Enter a general description for this model model: " + TermColors.ENDC)

                type_ = choose_option('model type', ['strategy', 'screener'],
                                      ["\t" + TermColors.BOLD + TermColors.WARNING +
                                       'A strategy is a model that uses blankly.Strategy' + TermColors.ENDC,
                                       "\t" + TermColors.BOLD + TermColors.WARNING +
                                       'A screener is a model uses blankly.Screener' + TermColors.ENDC])

                model_id = api.create_model(type_, model_name, general_description)['modelId']
            else:
                models = api.list_models()
                ids = []
                descriptions = []
                for i in models:
                    try:
                        id_ = i['id']
                        name_ = i['name']
                        ids.append(id_)
                        descriptions.append("\t" + TermColors.BOLD + TermColors.WARNING + id_ + ": " +
                                            TermColors.ENDC + TermColors.OKCYAN + name_)
                    except KeyError:
                        pass
                model_id = choose_option('model', ids, descriptions)

                # Because we have the ID, we now need to get the index directly
                index = None
                for i in models:
                    if i['id'] == model_id:
                        index = i

                if index is None:
                    raise LookupError("An issue occurred please try again.")

                type_ = index['type']
                model_name = index['name']

            info_print(f"Created a new model in blankly.json with ID: {model_id}")

            deployment_options['type'] = type_
            deployment_options['model_id'] = model_id
            deployment_options['model_name'] = model_name
        model_name = deployment_options['model_name']

        """
        This part generates API keys if they aren't found
        """
        if 'api_key' not in deployment_options or 'api_pass' not in deployment_options:
            response = api.generate_keys()
            deployment_options['api_key'] = response['apiKey']
            deployment_options['api_pass'] = response['apiPass']
            queue_write = True

        if queue_write:
            # Write the modified version with the ID back into the json file
            f = open(os.path.join(args['path'], deployment_script_name), 'w+')
            f.write(json.dumps(deployment_options, indent=2))
            f.close()
    except FileNotFoundError:
        raise FileNotFoundError(f"A {deployment_script_name} file must be present at the top level of the "
                                f"directory specified.")

    python_version = deployment_options['python_version']
    return model_name, deployment_options, python_version


temporary_zip_file = None


def zip_dir(args: dict, deployment_options: dict):
    global temporary_zip_file
    temporary_zip_file = tempfile.TemporaryDirectory()
    dist_directory = temporary_zip_file.__enter__()

    # dist_directory = tempfile.TemporaryDirectory().__enter__()
    # with tempfile.TemporaryDirectory() as dist_directory:
    source = os.path.abspath(args['path'])

    model_path = os.path.join(dist_directory, 'model.zip')
    zip_ = zipfile.ZipFile(model_path, 'w', zipfile.ZIP_DEFLATED)
    zipdir(source, zip_, deployment_options['ignore_files'])
    zip_.close()

    return model_path


def select_plan(api: API, plan_type: str):
    plans = api.get_plans(plan_type)

    plan_names = list(plans.keys())

    descriptions = []

    for i in plans:
        descriptions.append("\t" + TermColors.UNDERLINE + TermColors.OKBLUE + i + TermColors.ENDC +
                            "\n\t\t" + TermColors.OKGREEN + 'CPU: ' + str(plans[i]['cpu']) +
                            "\n\t\t" + TermColors.OKGREEN + 'RAM: ' + str(plans[i]['ram']) + TermColors.ENDC)

    chosen_plan = choose_option('plan', plan_names, descriptions)

    return chosen_plan


def add_path_arg(arg_parser, required=True):
    kwargs = {
        'metavar': 'path',
        'type': str,
        'help': 'Path to the directory containing the blankly enabled folder.',
        'nargs': '?'
    }
    # Pass this along if its not required
    if not required:
        kwargs['nargs'] = '?'
    arg_parser.add_argument('path', **kwargs)


def create_and_write_file(filename: str, default_contents: str = None):
    try:
        file = open("./" + filename, 'x+')
        if default_contents is not None:
            file.write(default_contents)
    except FileExistsError:
        print(f"{TermColors.WARNING}Already exists - skipping...{TermColors.ENDC}")


parser = argparse.ArgumentParser(description='Blankly CLI & deployment tool.')

subparsers = parser.add_subparsers(help='Different blankly commands.')

supported_exchanges_str = ''
for i in supported_exchanges:
    supported_exchanges_str += i + "\n"

init_parser = subparsers.add_parser('init', help=f'Sub command to create a blankly-enabled development environment.'
                                                 f' Supports blankly init on these exchanges: '
                                                 f'\n{supported_exchanges_str}- Ex: \'blankly init coinbase_pro\'')
init_parser.set_defaults(which='init')
# Generate a help message specific to this command
init_help_message = 'One of the following supported exchanges: \n'
for i in supported_exchanges:
    init_help_message += i + "\n"
init_parser.add_argument('exchange', nargs='?', default='none', type=str, help=init_help_message)

login_parser = subparsers.add_parser('login', help='Log in to your blankly account.')
login_parser.set_defaults(which='login')

login_parser = subparsers.add_parser('logout', help='Log out of your blankly account.')
login_parser.set_defaults(which='logout')

deploy_parser = subparsers.add_parser('deploy', help='Command to upload & start the model.')
deploy_parser.set_defaults(which='deploy')
add_path_arg(deploy_parser, required=False)

# Old project tools
# project_create_parser = subparsers.add_parser('create', help='Create a new project.')
# project_create_parser.set_defaults(which='create')
# list_parser = subparsers.add_parser('list', help='Show available projects & exit.')
# list_parser.set_defaults(which='list')

backtest_parser = subparsers.add_parser('backtest', help='Start a backtest on an uploaded model.')
backtest_parser.set_defaults(which='backtest')
add_path_arg(backtest_parser, required=False)


run_parser = subparsers.add_parser('run', help='Mimic the run mechanism used in blankly deployment.')
run_parser.add_argument('--monitor',
                        action='store_true',
                        help='Add this flag to monitor the blankly process to understand CPU & memory usage. '
                             'The module psutil must be installed.')
run_parser.set_defaults(which='run')
add_path_arg(run_parser)

# Create a global token value for use in the double nested function below
token = None
tokenfile_path = None


def __generate_tempfile():
    """
    Generate a temporary file with the blankly auth prefix and then get the directory / file name from it
    """
    fd, path = tempfile.mkstemp(prefix='blankly_auth_')
    temp_folder = os.path.dirname(path)
    file_name = os.path.basename(path)

    return fd, temp_folder, file_name


def logout():
    """
    This function will log out if not logged in
    """
    global tokenfile_path
    fd, temp_folder, file_name = __generate_tempfile()
    os.close(fd)
    os.remove(os.path.join(temp_folder, file_name))
    for i_ in os.listdir(temp_folder):
        # Check to see if one exists at this location
        if i_[0:13] == 'blankly_auth_' and i_ != file_name:
            # Kill the file we created
            os.remove(os.path.join(temp_folder, i_))


def is_logged_in():
    """
    This function will return if the user is logged in

    logged_in = __is_logged_in(temp_folder) is not None
    """
    global tokenfile_path
    fd, temp_folder, file_name = __generate_tempfile()
    for i_ in os.listdir(temp_folder):
        # Check to see if one exists at this location
        if i_[0:13] == 'blankly_auth_' and i_ != file_name:
            # Kill the file we created
            os.close(fd)
            os.remove(os.path.join(temp_folder, file_name))
            
            try:
                json.loads(open(temp_folder + '/' + i_).read())['token']
            except (KeyError, json.decoder.JSONDecodeError):
                return False

            # Cache the filepath globally
            tokenfile_path = os.path.join(temp_folder, i_)
            return True

    # Kill the file we created
    os.close(fd)
    os.remove(os.path.join(temp_folder, file_name))
    return False


def login(remove_cache: bool = False):
    # Set the token as global here
    global token
    # Also get the global path
    global tokenfile_path

    def load_token(token_path_: str) -> str:
        """
        If the file has the token then this function is a success
        """
        global tokenfile_path
        global token
        f_ = open(token_path_)
        token_file_ = json.load(f_)

        if 'token' in token_file_:
            # Globally cache the path
            tokenfile_path = token_path_
            # Globally cache the token
            token = token_file_['token']

            # Exit cleanly here finding the old refresh token
            return token_file_['token']

    # This can be skipped if is_logged_in is run which will find the temp folder and file name if necessary
    if tokenfile_path is None:
        fd, temp_folder, file_name = __generate_tempfile()
        for i_ in os.listdir(temp_folder):
            # Check to see if one exists at this location
            if i_[0:13] == 'blankly_auth_' and i_ != file_name:
                # If we're not removing cache this will use the old files to look for the token
                if not remove_cache:
                    # If it's different from the one that was just created, remove the one just created
                    os.close(fd)
                    os.remove(os.path.join(temp_folder, file_name))
                    # Reassign file name just in case its needed below to write into the file
                    # Note that we protect against corrupted files below by overwriting any contents in case
                    #  'token' not in token_file
                    file_name = i_
                    # Now just read the token from it
                    token_path = os.path.join(temp_folder, file_name)
                    try:
                        return load_token(token_path)
                    except json.decoder.JSONDecodeError:
                        # If it fails then don't return anything
                        #  just continue with re-logging in
                        pass
                # If we are removing cache then these files should just be deleted
                else:
                    os.remove(os.path.join(temp_folder, i_))
                # Be sure to leave the loop
                break
    elif isinstance(tokenfile_path, str):
        return load_token(tokenfile_path)

    from http.server import BaseHTTPRequestHandler, HTTPServer
    import urllib.parse

    class Handler(BaseHTTPRequestHandler):
        token: str

        def do_OPTIONS(self):
            # This options call is not used however these headers were hard to figure out, so I'm leaving them
            self.send_response(200, "ok")
            self.send_header('Access-Control-Allow-Origin', 'https://app.blankly.finance')
            self.send_header('Access-Control-Allow-Methods', 'GET, OPTIONS')
            self.send_header("Access-Control-Allow-Headers", 'Accept,Origin,Content-Type,X-LS-CORS-Template,'
                                                             'X-LS-Auth-Token,X-LS-Auth-User-Token,Content-Type,'
                                                             'X-LS-Sync-Result,X-LS-Sequence,token')
            self.end_headers()

        def do_GET(self):
            global token
            # Parse the URL
            args = urllib.parse.parse_qs(self.path[2:])
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()

            # Send this to the global static variable to use the info out of this
            if 'token' in args and args['token'] is not None:
                # Perform a GET request to pull down a successful response
                file = requests.get('https://firebasestorage.googleapis.com/v0/b/blankly-6ada5.appspot.com/o/'
                                    'login_success.html?alt=media&token=41d734e2-0a88-44c4-b1dd-7e081fd019e7')
                token = args['token'][0]

                info_print("Login success - You can close your browser.")
            else:
                file = requests.get('https://firebasestorage.googleapis.com/v0/b/blankly-6ada5.appspot.com/o/'
                                    'login_failure.html?alt=media&token=b2d9f6dc-aa54-4da2-aa6f-1f75a4025634')
                info_print("Failed to login - please try again.")

            # Write that back to the user
            self.wfile.write(bytes(file.text, "utf8"))

        def log_message(self, format_, *args):
            return

    server = HTTPServer(('', 9082), Handler)

    # Attempt to open the page
    url = 'https://app.blankly.finance/auth/signin?redirectUrl=/deploy'
    webbrowser.open_new(url)

    info_print('Log in on the pop-up window or navigate to: \nhttps://app.blankly.finance/auth/'
               'signin?redirectUrl=/deploy')

    # Continue waiting for GETs while the token is undefined
    while token is None:
        server.handle_request()

    cached_token_file = open(os.path.join(temp_folder, file_name), "w")
    json.dump({'token': token}, cached_token_file)

    return token


def zipdir(path, ziph, ignore_files: list):
    # From https://stackoverflow.com/a/1855118/8087739
    # Notice we're using this instead of shutil because it allows customization such as passwords and skipping
    # directories
    # ziph is zipfile handle

    filtered_ignore_files = []
    # Set all the ignored files to be absolute
    for i in range(len(ignore_files)):
        # Reject this case
        if ignore_files[i] == "":
            continue
        filtered_ignore_files.append(os.path.abspath(ignore_files[i]))

    for root, dirs, files in os.walk(path, topdown=True):
        for file in files:
            # (Modification) Skip everything that is in the blankly_dist folder
            filepath = os.path.join(root, file)

            if not (os.path.abspath(filepath) in filtered_ignore_files) and not (root in filtered_ignore_files):
                # This takes of the first part of the relative path and replaces it with /model/
                print(f'\tAdding: {file} in folder {root}.')
                relpath = os.path.relpath(filepath,
                                          os.path.join(path, '..'))
                relpath = os.path.normpath(relpath).split(os.sep)
                relpath[0] = os.sep + 'model'
                relpath = os.path.join(*relpath)

                ziph.write(filepath, relpath)
            else:
                info_print(f'\tSkipping: {file} in folder {root}.')


def main():
    args = vars(parser.parse_args())
    try:
        which = args['which']
    except KeyError:
        parser.print_help()
        return
    if which == 'deploy':
        token_ = login()

        api = API(token_)

        # Read and write to the deployment options if necessary
        model_name, deployment_options, python_version = \
            get_project_model_and_name(args, api)

        info_print("Zipping...")

        model_path = zip_dir(args, deployment_options)

        chosen_plan = select_plan(api, 'live')

        version_description = input(TermColors.BOLD + TermColors.WARNING +
                                    "Enter a description for this version of the model: " + TermColors.ENDC)

        info_print("Uploading...")
        kwargs = {
            'file_path': model_path,
            'model_id': deployment_options['model_id'],
            'version_description': version_description,
            'python_version': python_version,
            'type_': deployment_options['type'],
            'plan': chosen_plan
        }
        if kwargs['type_'] == 'screener':
            kwargs['schedule'] = deployment_options['screener']['schedule']

        response = api.deploy(**kwargs)
        if 'error' in response:
            info_print('Error: ' + response['error'])
        elif 'status' in response and response['status'] == 'success':
            info_print(f"Model upload completed at {response['timestamp']}:")
            info_print(f"\tModel ID:\t{response['modelId']}")
            info_print(f"\tVersion:\t{response['versionId']}")
            info_print(f"\tStatus:  \t{response['status']}")

    elif which == 'init':
        exchange = args['exchange']
        if exchange != 'none':
            user_defined_exchange = exchange.lower()
        else:
            user_defined_exchange = 'coinbase_pro'  # default to cbpro

        print("Initializing...")
        try:
            print(very_important_string)
        except UnicodeEncodeError:
            print("Welcome to blankly!")

        if user_defined_exchange not in supported_exchanges:
            base_str = 'Error: Please use one of our supported exchanges'
            exchanges = ', '.join(supported_exchanges)
            info_print(
                f'{base_str}: {TermColors.BOLD}{exchanges}{TermColors.ENDC}. '
                f'\nYour input: {TermColors.BOLD}{user_defined_exchange}{TermColors.ENDC}')
            return

        tld = 'com'
        if user_defined_exchange[0:7] == 'binance':
            # Find if it's .us if needed
            tld = user_defined_exchange[8:]
            user_defined_exchange = 'binance'

        exchange_config = {
            'settings.json': 'https://raw.githubusercontent.com/blankly-finance/examples/main/configs/settings.json',
            'keys.json': 'https://raw.githubusercontent.com/blankly-finance/examples/main/keys_example.json',
            'backtest_usd.json': 'https://raw.githubusercontent.com/blankly-finance/examples/main/configs/'
                                 'backtest_usd.json',
            'exchange': user_defined_exchange,
            'bot_url': f'https://raw.githubusercontent.com/blankly-finance/examples/main/init/rsi_'
                       f'{user_defined_exchange}.py',
            'tld': tld,
        }

        # Directly download keys.json
        print("Downloading keys template...")
        keys_example = requests.get(exchange_config['keys.json']).json()
        # No modification necessary
        create_and_write_file('keys.json', json.dumps(keys_example, indent=2))

        # Directly download settings.json
        print("Downloading settings defaults...")
        settings = requests.get(exchange_config['settings.json']).json()
        # Make sure we write the tld into the settings
        settings['settings']['binance']['binance_tld'] = exchange_config['tld']
        create_and_write_file('settings.json', json.dumps(settings, indent=2))

        # Directly download backtest.json
        print("Downloading backtest defaults...")
        backtest = requests.get(exchange_config['backtest_usd.json']).json()
        if user_defined_exchange == 'kucoin' or user_defined_exchange == 'binance':
            # USDT exchanges
            backtest['settings']['quote_account_value_in'] = 'USDT'
        create_and_write_file('backtest.json', json.dumps(backtest, indent=2))

        # Directly download a rsi bot
        print("Downloading RSI bot example...")
        bot = requests.get(exchange_config['bot_url'])
        create_and_write_file('bot.py', bot.text)

        print("Writing deployment defaults...")
        # Interpret defaults and write to this folder
        print("Detecting python version...")
        py_version = platform.python_version_tuple()
        print(f"{TermColors.OKCYAN}{TermColors.BOLD}Found python version: "
              f"{py_version[0]}.{py_version[1]}{TermColors.ENDC}")

        # Write in a blank requirements file
        print("Writing requirements.txt defaults...")
        create_and_write_file('requirements.txt', 'blankly')

        deploy = load_deployment_settings()

        deploy['python_version'] = py_version[0] + "." + py_version[1]
        create_and_write_file(deployment_script_name, json.dumps(deploy, indent=2))

        try:
            if is_logged_in():
                # We know we're logged in so make sure that we also get a project id and a model id
                print(f'{TermColors.WARNING}Automatically logged in!{TermColors.ENDC}')
                api = API(login())
                print(f'{TermColors.WARNING}Attaching this to a platform model...{TermColors.ENDC}')
                get_project_model_and_name(args, api)
            else:
                print(f'{TermColors.WARNING}Run \"blankly login\" and then \"blankly init\" again to get better '
                      f'backtest '
                      f'viewing.{TermColors.ENDC}')
        except Exception:
            traceback.print_exc()
            # Wipe the blankly.json to avoid confusion
            os.remove('./' + deployment_script_name)

        print(f"{TermColors.OKGREEN}{TermColors.UNDERLINE}Success!{TermColors.ENDC}")

    elif which == 'login':
        login(remove_cache=True)

    elif which == 'logout':
        logout()
        info_print("Cleared all blankly credentials.")

    # elif which == 'list':
    #     api = API(login())
    #     models = api.list_models()
    #     if len(models) > 0:
    #         print(TermColors.BOLD + TermColors.WARNING + "Models: " + TermColors.ENDC)
    #         for i in models:
    #             print("\t" + TermColors.BOLD + TermColors.WARNING + i['id'] + ": " +
    #                   TermColors.ENDC + TermColors.OKCYAN + i['name'])
    #             print(f"\t\t Description: {i['description']}")
    #     else:
    #         info_print("No projects found.")

    elif which == 'backtest':
        api = API(login())

        # Read and write to the deployment options if necessary
        model_name, deployment_options, python_version = get_project_model_and_name(args, api)

        if deployment_options['type'] == 'screener':
            raise AttributeError("Screeners are not backtestable.")

        info_print("Zipping...")

        model_path = zip_dir(args, deployment_options)

        chosen_plan = select_plan(api, 'backtesting')

        backtest_description = input(TermColors.BOLD + TermColors.WARNING +
                                     "Enter a backtest description for this version of the model: " + TermColors.ENDC)
        info_print("Uploading...")

        response = api.backtest(file_path=model_path,
                                model_id=deployment_options['model_id'],
                                args=deployment_options['backtest_args'],
                                plan=chosen_plan,
                                python_version=python_version,
                                backtest_description=backtest_description,
                                type_=deployment_options['type'])

        info_print("Uploading...")
        if 'error' in response:
            info_print(response['error'])
        else:
            info_print(f"Backtest upload completed at {response['timestamp']}:")
            info_print(f"\tModel ID:\t{response['modelId']}")
            info_print(f"\tVersion:\t{response['versionId']}")
            info_print(f"\tStatus:  \t{response['status']}")

    # elif which == 'create':
    #     api = API(login())
    #
    #     project_name = input(TermColors.BOLD + "Input a name for your model: " + TermColors.ENDC)
    #
    #     project_description = input(TermColors.BOLD + "Input a description for your model: " + TermColors.ENDC)
    #
    #     input(TermColors.BOLD + TermColors.FAIL + "Press enter to confirm model creation." + TermColors.ENDC)
    #
    #     response = api.create_model(type_='strategy', name=project_name, description=project_description)
    #
    #     if 'error' in response:
    #         info_print('Error: ' + response['error'])
    #     elif 'status' in response and response['status'] == 'success':
    #         info_print(f"Created {response['name']} at {response['createdAt']}:")
    #         info_print(f"\tModel Id:\t{response['id']}")

    elif which == 'run':
        if args['path'] is None:
            run_parser.print_help()
        else:
            current_directory = os.getcwd()
            blankly_folder = os.path.join(current_directory, args['path'])
            deployment_location = os.path.join(
                blankly_folder, deployment_script_name)
            try:
                # Find where the user specified their working directory and migrate to that
                deployment_dict = load_json_file(deployment_location)
            except FileNotFoundError:
                raise FileNotFoundError(f"Deployment json not found in location {deployment_location}.")

            # Set the working directory to match the deployment dictionary
            os.chdir(os.path.join(blankly_folder, deployment_dict['working_directory']))

            current_version = platform.python_version_tuple()
            current_version = current_version[0] + "." + current_version[1]

            if current_version != str(deployment_dict['python_version']):
                warn_string = f"Specified python version different than current interpreter. Using version " \
                              f"{platform.python_version()}. The specified version will be followed on full" \
                              f" deployment."
                warnings.warn(warn_string)

            if 'requirements' in deployment_dict and deployment_dict['requirements'] is not None:
                warning_string = "Requirements specified but not installed in this test script. Install " \
                                 "manually if needed."
                info_print(warning_string)

            sys.path.append(blankly_folder)

            # The execute function to run the modules
            main_script_abs = os.path.abspath(deployment_dict['main_script'])
            runpy.run_path(main_script_abs, {}, "__main__")

            if args['monitor']:
                try:
                    import psutil
                except ImportError:
                    raise ImportError("Must install psutil (pip install psutil) to monitor "
                                      "the Blankly process")
                print("Process must be killed to halt monitoring.")

                pid = os.getpid()
                process = psutil.Process(pid)
                while True:
                    monitor_string = "\rCPU Usage: "
                    monitor_string += str(process.cpu_percent())
                    monitor_string += " | "
                    monitor_string += "Memory Usage: "
                    monitor_string += str(process.memory_percent())
                    sys.stdout.write(monitor_string)
                    sys.stdout.flush()
                    time.sleep(5)


if __name__ == "__main__":
    main()



"""
    Timezone & time utils
    Copyright (C) 2021 Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from datetime import datetime as dt


def is_datetime_naive(datetime_obj: dt):
    if datetime_obj.tzinfo is None or datetime_obj.tzinfo.utcoffset(datetime_obj) is None:
        return True
    return False



"""
    Base FuturesExchange class.
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import abc
import time

import blankly
from blankly.exchanges.abc_base_exchange import ABCBaseExchange
from blankly.exchanges.auth.utils import write_auth_cache
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface


class FuturesExchange(ABCBaseExchange, abc.ABC):
    exchange_type: str
    portfolio_name: str

    def __init__(self, exchange_type, portfolio_name, preferences_path):
        self.exchange_type = exchange_type  # binance, ftx
        self.portfolio_name = portfolio_name  # my_cool_portfolio
        self.preferences = blankly.utils.load_user_preferences(
            preferences_path)
        self.models = {}

    @property
    @abc.abstractmethod
    def calls(self):
        """Authenticated direct calls object"""
        pass

    @property
    @abc.abstractmethod
    def interface(self) -> FuturesExchangeInterface:
        """This FuturesExchange's Interface"""
        pass

    def initialize(self):
        """
        Initialize the exchange. This should usually be called at the end of subclass constructors.
        This will report the exchange type to blankly.reporter and initialize the cache for this exchange.
        """
        blankly.reporter.export_used_exchange(self.exchange_type)
        write_auth_cache(self.exchange_type, self.portfolio_name, self.calls)

    def get_name(self) -> str:
        return self.portfolio_name

    def get_type(self) -> str:
        return self.exchange_type

    def get_interface(self):
        return self.interface



"""
    Logging system to allow users to view & understand actions done in the strategy
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pandas
from typing import Union
from datetime import datetime as dt

import blankly
from blankly.exchanges.interfaces.abc_exchange_interface import ABCExchangeInterface
from blankly.exchanges.orders.market_order import MarketOrder
from blankly.exchanges.orders.limit_order import LimitOrder
from blankly.utils.utils import AttributeDict


class StrategyLogger(ABCExchangeInterface):
    def __init__(self, interface=None, strategy=None):
        self.interface = interface
        self.strategy = strategy
        self.__type = self.interface.get_exchange_type()

    def get_calls(self):
        """
        No logging implemented
        """
        return self.interface.get_calls()
    
    def get_exchange_type(self):
        """
        No logging implemented
        """
        return self.interface.get_exchange_type()
    
    def get_account(self, symbol: str = None) -> AttributeDict:
        """
        No logging implemented
        """
        return self.interface.get_account(symbol)

    """
    These next three queries have large responses. It is unclear if this quantity of data is useful or necessary
    """
    def get_products(self):
        """
        No logging implemented
        """
        return self.interface.get_products()

    def get_product_history(self, symbol: str, epoch_start: float,
                            epoch_stop: float, resolution: Union[str, int]) -> pandas.DataFrame:
        """
        No logging implemented
        """
        return self.interface.get_product_history(symbol, epoch_start, epoch_stop, resolution)
    
    def history(self,
                symbol: str,
                to: Union[str, int] = 200,
                resolution: Union[str, int] = '1d',
                start_date: Union[str, dt, float] = None,
                end_date: Union[str, dt, float] = None,
                return_as: str = 'df') -> pandas.DataFrame:
        """
        No logging implemented
        """
        return self.interface.history(symbol, to=to,
                                      resolution=resolution, start_date=start_date,
                                      end_date=end_date, return_as=return_as)
    
    def market_order(self, symbol: str, side: str, size: float) -> MarketOrder:
        out = self.interface.market_order(symbol, side, size)

        # Record this market order along with the arguments
        try:
            blankly.reporter.log_market_order({
                'symbol': symbol,
                'exchange': self.__type,
                'size': size,
                'id': out.get_id(),
                'side': side
            }, self.__type)
        except Exception:
            pass
        return out
    
    def limit_order(self, symbol: str, side: str, price: float, size: float) -> LimitOrder:
        out = self.interface.limit_order(symbol, side, price, size)

        # Record limit order along with the arguments
        try:
            blankly.reporter.log_limit_order({
                'symbol': symbol,
                'exchange': self.__type,
                'size': size,
                'id': out.get_id(),
                'side': side,
                'price': price
            }, self.__type)
        except Exception:
            pass
        return out
    
    def cancel_order(self, symbol: str, order_id: str) -> dict:
        """
        No logging implemented
        """
        return self.interface.cancel_order(symbol, order_id)

    def get_open_orders(self, symbol: str = None) -> list:
        """
        No logging implemented
        """
        return self.interface.get_open_orders(symbol=symbol)

    def get_order(self, symbol: str, order_id: str) -> dict:
        """
        TODO - this needs to update the order on the backend
        """
        out = self.interface.get_order(symbol, order_id)

        # Record the arguments
        try:
            blankly.reporter.update_order({
                'id': order_id,
                'status': out['status']
            }, self.__type)
        except Exception:
            pass
        return out

    def get_fees(self) -> dict:
        """
        No logging implemented
        """
        return self.interface.get_fees()

    def get_order_filter(self, symbol: str):
        """
        No logging implemented
        """
        return self.interface.get_order_filter(symbol)
    
    def get_price(self, symbol: str) -> float:
        """
        No logging implemented
        """
        return self.interface.get_price(symbol)

    """
    No logging implemented for these properties
    """
    @property
    def account(self) -> AttributeDict:
        """
        No logging implemented
        """
        return self.interface.account
    
    @property
    def orders(self) -> list:
        """
        No logging implemented
        """
        return self.interface.orders

    @property
    def cash(self) -> float:
        """
        No logging implemented
        """
        return self.interface.cash
    



"""
    Base authentication for kucoin
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import time

import pandas as pd

import blankly.utils.time_builder
import blankly.utils.utils as utils
from blankly.exchanges.interfaces.exchange_interface import ExchangeInterface
from blankly.exchanges.orders.limit_order import LimitOrder
from blankly.exchanges.orders.market_order import MarketOrder
from blankly.utils.exceptions import APIException, InvalidOrder


class KucoinInterface(ExchangeInterface):
    def __init__(self, exchange_name, authenticated_api):

        super().__init__(exchange_name, authenticated_api, valid_resolutions=[60, 180, 300, 900, 1800,
                                                                              3600, 7200, 14400, 21600,
                                                                              28800, 43200, 86400, 604800])

        try:
            from kucoin import client as KucoinAPI
        except ImportError:
            raise ImportError("Please \"pip install kucoin-python\" to use kucoin with blankly.")

        self._market: KucoinAPI.Market = self.calls['market']
        self._trade: KucoinAPI.Trade = self.calls['trade']
        self._user: KucoinAPI.User = self.calls['user']

    def init_exchange(self):
        fees = self.calls['user'].get_base_fee()
        try:
            if fees['msg'] == "Invalid API Key":
                raise LookupError("Invalid API Key - are you trying to use your normal exchange keys "
                                  "while in sandbox mode? \nTry toggling the \'sandbox\' setting "
                                  "in your keys.json or check if the keys were input correctly into your "
                                  "keys.json.")
        except KeyError:
            pass

    def get_products(self):
        needed = self.needed['get_products']
        """
        [
          {
            "symbol": "BTC-USDT",
            "name": "BTC-USDT",
            "baseCurrency": "BTC",
            "quoteCurrency": "USDT",
            "baseMinSize": "0.00000001",
            "quoteMinSize": "0.01",
            "baseMaxSize": "10000",
            "quoteMaxSize": "100000",
            "baseIncrement": "0.00000001",
            "quoteIncrement": "0.01",
            "priceIncrement": "0.00000001",
            "feeCurrency": "USDT",
            "enableTrading": true,
            "isMarginEnabled": true,
            "priceLimitRate": "0.1"
          }
        ]
        """
        products = self._market.get_symbol_list()
        for i in range(len(products)):
            products[i]["base_asset"] = products[i].pop("baseCurrency")
            products[i]["quote_asset"] = products[i].pop("quoteCurrency")
            products[i]["base_min_size"] = products[i].pop("baseMinSize")
            products[i]["base_max_size"] = products[i].pop("baseMaxSize")
            products[i]["base_increment"] = products[i].pop("baseIncrement")
            products[i] = utils.isolate_specific(needed, products[i])
        return products

    @utils.enforce_base_asset
    def get_account(self, symbol=None) -> utils.AttributeDict:
        """
            Get all currencies in an account, or sort by symbol
            Args:
                symbol (Optional): Filter by particular symbol
                These arguments are mutually exclusive
        """
        symbol = super().get_account(symbol=symbol)

        """
        [
              {
                "id": "5bd6e9286d99522a52e458de",  //accountId
                "currency": "BTC",  //Currency
                "type": "main",     //Account type, including main and trade
                "balance": "237582.04299",  //Total assets of a currency
                "available": "237582.032",  //Available assets of a currency
                "holds": "0.01099" //Hold assets of a currency
              },
              {
                ...
              }
        """
        accounts = self._user.get_account_list()
        trade_accounts = []
        # Filter for only the trade accounts
        for i in accounts:
            if i['type'] == 'trade':
                trade_accounts.append(i)

        # If we have no positions this is a case that things can go wrong
        if isinstance(trade_accounts, dict):
            trade_accounts = []

        # We have to sort through it if the accounts are none
        if symbol is not None:
            for i in trade_accounts:
                if i["currency"] == symbol:
                    return utils.AttributeDict({
                        'available': float(i['available']),
                        'hold': float(i['holds'])
                    })

        parsed_dictionary = utils.AttributeDict({})

        # If the first loop didn't find it we'll try again here
        for i in range(len(trade_accounts)):
            parsed_dictionary[trade_accounts[i]['currency']] = utils.AttributeDict({
                'available': float(trade_accounts[i]['available']),
                'hold': float(trade_accounts[i]['holds'])
            })

        parsed_dictionary = utils.add_all_products(parsed_dictionary, self.get_products())

        if symbol is not None:
            if symbol in parsed_dictionary:
                return parsed_dictionary[symbol]
            else:
                raise KeyError(f'Could not find account for asset {symbol}')

        return parsed_dictionary

    @utils.order_protection
    def market_order(self, symbol, side, size) -> MarketOrder:
        """
            Used for buying or selling market orders
            Args:
                symbol: currency to buy
                side: buy/sell
                size: desired amount of base currency to buy or sell
        """
        needed = self.needed['market_order']
        """
        {
            "orderId": "5bd6e9286d99522a52e458de"
        }
        """
        order = {
            'symbol': symbol,
            'side': side,
            'size': size,
            'type': 'market',
        }

        response = self._trade.create_market_order(symbol, side, size=size)
        if "msg" in response:
            raise InvalidOrder("Invalid Order: " + response["msg"])

        response_details = self._trade.get_order_details(response['orderId'])
        """
        {
            "id": "5c35c02703aa673ceec2a168", //
            "symbol": "BTC-USDT", //
            "opType": "DEAL",
            "type": "limit", //
            "side": "buy", //
            "price": "10",
            "size": "2", //
            "funds": "0",
            "dealFunds": "0.166",
            "dealSize": "2",
            "fee": "0",
            "feeCurrency": "USDT",
            "stp": "",
            "stop": "",
            "stopTriggered": false,
            "stopPrice": "0",
            "timeInForce": "GTC",
            "postOnly": false,
            "hidden": false,
            "iceberg": false,
            "visibleSize": "0",
            "cancelAfter": 0,
            "channel": "IOS",
            "clientOid": "",
            "remark": "",
            "tags": "",
            "isActive": false, //
            "cancelExist": false,
            "createdAt": 1547026471000, //
            "tradeType": "TRADE"
        }
        """

        response_details["created_at"] = response_details.pop('createdAt')
        response_details["status"] = response_details.pop('isActive')
        response_details = utils.isolate_specific(needed, response_details)
        return MarketOrder(order, response_details, self)

    @utils.order_protection
    def limit_order(self, symbol, side, price, size) -> LimitOrder:
        """
            Used for buying or selling limit orders
            Args:
                symbol: currency to buy
                side: buy/sell
                price: price to set limit order
                size: amount of currency (like BTC) for the limit to be valued
        """
        needed = self.needed['limit_order']

        order = {
            'symbol': symbol,
            'side': side,
            'size': size,
            'price': price,
            'type': 'limit',
        }
        response = self._trade.create_limit_order(symbol, side, size, price)
        if "msg" in response:
            raise InvalidOrder("Invalid Order: " + response["msg"])

        response_details = self._trade.get_order_details(response['orderId'])
        """
               {
                   "id": "5c35c02703aa673ceec2a168", //
                   "symbol": "BTC-USDT", //
                   "opType": "DEAL",
                   "type": "limit", //
                   "side": "buy", //
                   "price": "10",
                   "size": "2", //
                   "funds": "0",
                   "dealFunds": "0.166",
                   "dealSize": "2",
                   "fee": "0",
                   "feeCurrency": "USDT",
                   "stp": "",
                   "stop": "",
                   "stopTriggered": false,
                   "stopPrice": "0",
                   "timeInForce": "GTC", //
                   "postOnly": false,
                   "hidden": false,
                   "iceberg": false,
                   "visibleSize": "0",
                   "cancelAfter": 0,
                   "channel": "IOS",
                   "clientOid": "",
                   "remark": "",
                   "tags": "",
                   "isActive": false, //
                   "cancelExist": false,
                   "createdAt": 1547026471000, //
                   "tradeType": "TRADE"
               }
               """

        response_details["created_at"] = response_details.pop('createdAt')
        response_details["time_in_force"] = response_details.pop('timeInForce')
        response_details["status"] = response_details.pop('isActive')
        if response_details["status"]:
            response_details["status"] = 'pending'
        else:
            response_details["status"] = 'done'
        response_details = utils.isolate_specific(needed, response_details)
        return LimitOrder(order, response_details, self)

    def cancel_order(self, symbol, order_id) -> dict:
        """
        Returns:
           dict: Containing the order_id of cancelled order. Example::
           { "cancelledOrderIds": ["c5ab5eae-76be-480e-8961-00792dc7e138"]}
        """
        return {"order_id": self._trade.cancel_order(order_id)['cancelledOrderIds'][0]}

    def get_open_orders(self, symbol=None):
        """
        List open orders.
        """
        """
        {
            "currentPage": 1,
            "pageSize": 1,
            "totalNum": 153408,
            "totalPage": 153408,
            "items": [
              {
                "id": "5c35c02703aa673ceec2a168",   //orderid
                "symbol": "BTC-USDT",   //symbol
                "opType": "DEAL",      // operation type: DEAL
                "type": "limit",       // order type,e.g. limit,market,stop_limit.
                "side": "buy",         // transaction direction,include buy and sell
                "price": "10",         // order price
                "size": "2",           // order quantity
                "funds": "0",          // order funds
                "dealFunds": "0.166",  // deal funds
                "dealSize": "2",       // deal quantity
                "fee": "0",            // fee
                "feeCurrency": "USDT", // charge fee currency
                "stp": "",             // self trade prevention,include CN,CO,DC,CB
                "stop": "",            // stop type
                "stopTriggered": false,  // stop order is triggered
                "stopPrice": "0",      // stop price
                "timeInForce": "GTC",  // time InForce,include GTC,GTT,IOC,FOK
                "postOnly": false,     // postOnly
                "hidden": false,       // hidden order
                "iceberg": false,      // iceberg order
                "visibleSize": "0",    // display quantity for iceberg order
                "cancelAfter": 0,      // cancel orders time，requires timeInForce to be GTT
                "channel": "IOS",      // order source
                "clientOid": "",       // user-entered order unique mark
                "remark": "",          // remark
                "tags": "",            // tag order source
                "isActive": false,     // status before unfilled or uncancelled
                "cancelExist": false,   // order cancellation transaction record
                "createdAt": 1547026471000,  // create time
                "tradeType": "TRADE"
              }
            ]
         }
        """
        if symbol is None:
            open_orders = list(self._trade.get_order_list(status='active')["items"])
        else:
            open_orders = list(self._trade.get_order_list(status='active', symbol=symbol)["items"])

        if len(open_orders) == 0:
            return []
        if open_orders[0] == 'msg':
            raise APIException("Error: " + str(open_orders))

        for i in range(len(open_orders)):
            # open_orders[i]["createdAt"] = utils.epoch_from_ISO8601(open_orders[i]["createdAt"])
            needed = self.choose_order_specificity(open_orders[i]['type'])
            open_orders[i]["time_in_force"] = open_orders[i].pop('timeInForce')
            open_orders[i]["created_at"] = open_orders[i].pop('createdAt')
            open_orders[i]["status"] = open_orders[i].pop('isActive')
            if open_orders[i]["status"]:
                open_orders[i]["status"] = 'pending'
            else:
                open_orders[i]["status"] = 'done'

            open_orders[i] = utils.isolate_specific(needed, open_orders[i])

        return open_orders

    def get_order(self, symbol, order_id) -> dict:
        """
        {
            "id": "5c35c02703aa673ceec2a168",
            "symbol": "BTC-USDT",
            "opType": "DEAL",
            "type": "limit",
            "side": "buy",
            "price": "10",
            "size": "2",
            "funds": "0",
            "dealFunds": "0.166",
            "dealSize": "2",
            "fee": "0",
            "feeCurrency": "USDT",
            "stp": "",
            "stop": "",
            "stopTriggered": false,
            "stopPrice": "0",
            "timeInForce": "GTC",
            "postOnly": false,
            "hidden": false,
            "iceberg": false,
            "visibleSize": "0",
            "cancelAfter": 0,
            "channel": "IOS",
            "clientOid": "",
            "remark": "",
            "tags": "",
            "isActive": false,
            "cancelExist": false,
            "createdAt": 1547026471000,
            "tradeType": "TRADE"
        }
        """
        response = self._trade.get_order_details(order_id)

        if 'msg' in response:
            raise APIException("Invalid: " + str(response['msg']) + ", was the order canceled?")

        if isinstance(response['createdAt'], str):
            response["createdAt"] = utils.epoch_from_iso8601(response["createdAt"])

        if response['type'] == 'market':
            needed = self.needed['market_order']
        elif response['type'] == 'limit':
            needed = self.needed['limit_order']
        else:
            needed = self.needed['market_order']

        response["created_at"] = response.pop('createdAt')
        response["status"] = response.pop('isActive')
        if response["status"]:
            response["status"] = 'pending'
        else:
            response["status"] = 'done'
        response["time_in_force"] = response.pop('timeInForce')

        return utils.isolate_specific(needed, response)

    def get_fees(self) -> dict:
        needed = self.needed['get_fees']
        """
            {
                "code": "200000",
                "data": {
                    "takerFeeRate": "0.001",
                    "makerFeeRate": "0.001"
                }
            }
        """
        fees = self._user.get_base_fee()
        fees["taker_fee_rate"] = fees.pop('takerFeeRate')
        fees["maker_fee_rate"] = fees.pop('makerFeeRate')
        return utils.isolate_specific(needed, fees)

    def overridden_history(self, symbol, epoch_start, epoch_stop, resolution, **kwargs) -> pd.DataFrame:
        """
        Kucoin is strange because it's exclusive instead of inclusive. This generally invovles adding an extra
        datapoint so this is here to do some of that work
        """
        to = kwargs['to']

        if isinstance(to, str):
            epoch_start -= resolution
        elif isinstance(to, int):
            epoch_start = epoch_stop - (to * resolution)

        return self.get_product_history(symbol, epoch_start, epoch_stop, resolution)

    def get_product_history(self, symbol, epoch_start, epoch_stop, resolution):
        """
        Returns the product history from an exchange
        Args:
            symbol: blankly product ID format (BTC-USD)
            epoch_start: Time to begin download
            epoch_stop: Time to stop download
            resolution: Resolution in seconds between tick (ex: 60 = 1 per minute)
        Returns:
            Dataframe with *at least* 'time (epoch)', 'low', 'high', 'open', 'close', 'volume' as columns.
        """
        resolution = blankly.time_builder.time_interval_to_seconds(resolution)

        epoch_start = int(utils.convert_epochs(epoch_start))
        epoch_stop = int(utils.convert_epochs(epoch_stop))

        accepted_grans = [60, 180, 300, 900, 1800, 3600, 7200, 14400, 21600, 28800, 43200, 86400, 604800]

        if resolution not in accepted_grans:
            utils.info_print("Granularity is not an accepted granularity...rounding to nearest valid value.")
            resolution = accepted_grans[min(range(len(accepted_grans)),
                                            key=lambda i: abs(accepted_grans[i] - resolution))]

        resolution = int(resolution)

        epoch_start -= resolution

        lookup_dict = {
            60: '1min',
            180: '3min',
            300: '5min',
            900: '15min',
            1800: '30min',
            3600: '1hour',
            7200: '2hour',
            14400: '4hour',
            21600: '6hour',
            28800: '8hour',
            43200: '12hour',
            86400: '1day',
            604800: '1week',
        }
        gran_string = lookup_dict[resolution]

        # Figure out how many points are needed
        need = int((epoch_stop - epoch_start) / resolution)
        initial_need = need
        window_open = epoch_start
        history = []

        while need > 1500:
            # Close is always 300 points ahead
            window_close = int(window_open + 1500 * resolution)
            response = self._market.get_kline(symbol, gran_string,
                                              startAt=window_open, endAt=window_close)
            if isinstance(response, dict):
                raise APIException(response['msg'])
            history = history + response

            window_open = window_close
            need -= 1500
            time.sleep(2)
            utils.update_progress((initial_need - need) / initial_need)

        # Fill the remainder
        response = self._market.get_kline(symbol, gran_string,
                                          startAt=window_open, endAt=epoch_stop)
        if isinstance(response, dict):
            raise APIException(response['msg'])

        history = history + response
        history.sort(key=lambda x: x[0])

        df = pd.DataFrame(history, columns=['time', 'open', 'close', 'high', 'low', 'volume', 'turnover'])
        del df['turnover']

        # Have to cast this for some reason
        # df[['time']] = df[['time']].astype(float)
        df[['time']] = df[['time']].astype(int)
        df[['open', 'close', 'high', 'low', 'volume']] = df[['open', 'close', 'high', 'low', 'volume']].astype(float)

        try:
            epoch_start += resolution
            df = (df.iloc[int(-(epoch_stop-epoch_start)/resolution):]).reset_index(drop=True)
        except Exception:
            pass

        return df.reindex(columns=['time', 'low', 'high', 'open', 'close', 'volume'])

    def get_order_filter(self, symbol: str):
        """
        Returns:
        list: List of available currency pairs. Example::
           [
              {
                "symbol": "BTC-USDT",
                "name": "BTC-USDT",
                "baseCurrency": "BTC",
                "quoteCurrency": "USDT",
                "baseMinSize": "0.00000001",
                "quoteMinSize": "0.01",
                "baseMaxSize": "10000",
                "quoteMaxSize": "100000",
                "baseIncrement": "0.00000001",
                "quoteIncrement": "0.01",
                "priceIncrement": "0.00000001",
                "feeCurrency": "USDT",
                "enableTrading": true,
                "isMarginEnabled": true,
                "priceLimitRate": "0.1"
              }
           ]
        """
        response = self._market.get_symbol_list()
        products = None
        for i in response:
            if i["symbol"] == symbol:
                products = i
                break

        if products is None:
            raise LookupError("Specified market not found")

        base_min_size = float(products.pop('baseMinSize'))
        base_max_size = float(products.pop('baseMaxSize'))
        base_increment = float(products.pop('baseIncrement'))

        return {
            "symbol": products.pop('symbol'),
            "base_asset": products.pop('baseCurrency'),
            "quote_asset": products.pop('quoteCurrency'),
            "max_orders": 1000000000000,
            "limit_order": {
                "base_min_size": base_min_size,  # Minimum size to buy
                "base_max_size": base_max_size,  # Maximum size to buy
                "base_increment": base_increment,  # Specifies the minimum increment
                # for the base_asset.
                "price_increment": float(products['quoteIncrement']),

                "min_price": float(products['quoteIncrement']),
                "max_price": 9999999999,
            },
            'market_order': {
                "fractionable": True,

                "base_min_size": base_min_size,  # Minimum size to buy
                "base_max_size": base_max_size,  # Maximum size to buy
                "base_increment": base_increment,  # Specifies the minimum increment

                "quote_increment": float(products.pop('quoteIncrement')),  # Specifies the min order price as well
                # as the price increment.
                "buy": {
                    "min_funds": float(products['quoteMinSize']),
                    "max_funds": float(products['quoteMaxSize']),
                },
                "sell": {
                    "min_funds": float(products.pop('quoteMinSize')),
                    "max_funds": float(products.pop('quoteMaxSize')),
                },
            },
            "exchange_specific": {**products}
        }

    def get_price(self, symbol) -> float:
        """
        Returns the best bid price and size,
        the best ask price and size as well
        as the last traded price and the last traded size
        """
        """
         {
            "sequence": "1550467636704",
            "bestAsk": "0.03715004",
            "size": "0.17",
            "price": "0.03715005",
            "bestBidSize": "3.803",
            "bestBid": "0.03710768",
            "bestAskSize": "1.788",
            "time": 1550653727731
        }       
        """
        response = self._market.get_ticker(symbol)
        if 'msg' in response:
            raise APIException("Error: " + response['msg'])
        return float(response['price'])



from typing import List


def get_open_order_ids(open_orders: List[dict]) -> List[str]:

    ids: List[str] = []

    for order in open_orders:
        ids.append(str(order["id"]))
    
    return ids



"""
    Inherited exchange object.
    Copyright (C) 2021  Emerson Dove, Arun Annamalai, Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import abc
import time

import blankly
from blankly.exchanges.abc_exchange import ABCExchange
from blankly.exchanges.auth.utils import write_auth_cache
from blankly.exchanges.interfaces.abc_exchange_interface import ABCExchangeInterface
from blankly.exchanges.interfaces.coinbase_pro.coinbase_pro_interface import CoinbaseProInterface
from blankly.exchanges.interfaces.oanda.oanda_interface import OandaInterface
from blankly.exchanges.interfaces.ftx.ftx_interface import FTXInterface
from blankly.exchanges.interfaces.alpaca.alpaca_interface import AlpacaInterface
from blankly.exchanges.interfaces.binance.binance_interface import BinanceInterface
from blankly.exchanges.interfaces.kucoin.kucoin_interface import KucoinInterface


class Exchange(ABCExchange, abc.ABC):
    interface: ABCExchangeInterface

    def __init__(self, exchange_type, portfolio_name, preferences_path):
        self.__type = exchange_type  # coinbase_pro, binance, alpaca, oanda, ftx
        self.__name = portfolio_name  # my_cool_portfolio

        # Make a public version of portfolio name
        self.portfolio_name = self.__name

        self.preferences = blankly.utils.load_user_preferences(preferences_path)

        self.models = {}

        # Fill this in the method below
        self.calls = None

    # TODO this will be removed in the next update
    def evaluate_sandbox(self, auth):
        """
        This will try to maintain compatibility with older versions if they fail to pivot to the new version immediately
        """
        if 'sandbox' not in auth.keys:
            try:
                return self.preferences['settings']['use_sandbox']
            except KeyError:
                raise KeyError("No sandbox setting found in either settings.json or keys.json. Please use the example"
                               " above this error to modify your keys.json.")
        else:
            return auth.keys['sandbox']

    def construct_interface_and_cache(self, calls):
        """
        If you are a contributor, you need to modify this function to add exchanges
        The core functions that creates the interface based on the exchange type & automatically caches
        """
        self.calls = calls
        if self.__type == "coinbase_pro":
            self.interface = CoinbaseProInterface(self.__type, calls)
        elif self.__type == "binance":
            self.interface = BinanceInterface(self.__type, calls)
        elif self.__type == "alpaca":
            self.interface = AlpacaInterface(self.__type, calls)
        elif self.__type == "ftx":
            self.interface = FTXInterface(self.__type, calls)
        elif self.__type == "oanda":
            self.interface = OandaInterface(self.__type, calls)
        elif self.__type == "kucoin":
            self.interface = KucoinInterface(self.__type, calls)

        blankly.reporter.export_used_exchange(self.__type)

        write_auth_cache(self.__type, self.__name, calls)

    def get_name(self):
        return self.__name

    def get_type(self):
        return self.__type

    def get_preferences(self):
        return self.preferences

    def get_interface(self) -> ABCExchangeInterface:
        """
        Get the the authenticated interface for the object. This will provide authenticated API calls.
        """
        return self.interface

    def start_models(self, symbol=None):
        """
        Start all models or a specific one after appending it to to the exchange.

        This is used only for multiprocessed bots which are appended directly to the exchange. NOT bots that use
        the strategy class.
        """
        if symbol is not None:
            # Run a specific model with the args
            if not self.models[symbol]["model"].is_running():
                self.models[symbol]["model"].run(self.models[symbol]["args"])
        else:
            for coin_iterator in self.models:
                # Start all models
                if not self.models[coin_iterator]["model"].is_running():
                    self.models[coin_iterator]["model"].run(self.models[coin_iterator]["args"])
                    # There is some delay. Optimally the bots should be started in a different thread so this doesn't
                    # block the main
                    time.sleep(.1)
                else:
                    print("Ignoring the model on " + coin_iterator)

    def __get_model(self, coin):
        """
        Get the model to reference to
        """
        return self.models[coin]["model"]

    def get_model_state(self, symbol):
        """
        Returns JUST the model state, as opposed to all the data returned by get_asset_state()

        This is also used for querying a general state from a multiprocessed bot
        Args:
            symbol: Currency that the selected model is running on.
        """
        return (self.__get_model(symbol)).get_state()

    def get_full_state(self, symbol):
        """
        Makes API calls to determine the state of the currency. This also returns the state of the model on that
        currency.

        This is also used only for multiprocessed models

        Args:
            symbol: Currency to filter for. This filters model information and the exchange information.
        """
        state = self.get_asset_state(symbol)

        return {
            "account": state,
            "model": self.get_model_state(symbol)
        }

    def write_value(self, symbol, key, value):
        """
        Write a key/value pair to a bot attached to a particular currency pair

        This is also used only for multiprocessed bots.

        Args:
            symbol: Change state on bot attached to this currency
            key: Key to assign a value to
            value: Value to assign to the key
        """
        self.__get_model(symbol).update_state(key, value)

    def append_model(self, model, symbol, args=None):
        """
        Append the models to the exchange, these can be run
        Args:
            model: Model object to be used. This is objects inheriting blankly_bot
            symbol: the currency to use, such as "BTC-USD"
            args: Args to pass into the model when it is run. This can be any datatype, the object is passed
        """
        added_model = model
        self.models[symbol] = {
            "model": added_model,
            "args": args
        }
        model.setup(self.__type, symbol, self.preferences, self.get_full_state(symbol),
                    self.interface)

    @abc.abstractmethod
    def get_asset_state(self, symbol):
        """
        This is left out of documentation because it lacks thorough implementations.

        Generally this will be used to get a report of performance of a symbol on an exchange.
        """
        pass

    @abc.abstractmethod
    def get_direct_calls(self):
        pass
